/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use std::cell::RefCell;
use std::sync::atomic::{AtomicBool, Ordering};

use color_eyre::eyre::{Result, WrapErr};
use crossterm::event::KeyEvent;
use futures::stream::FuturesUnordered;
use futures::StreamExt;
use nonempty::NonEmpty;
use ratatui::prelude::Rect;
use serde::{Deserialize, Serialize};
use tokio::sync::mpsc;

use crate::{
  action::Action,
  commands::RataCommands,
  components::{fps::FpsCounter, home::Home, Component},
  config::Config,
  mode::Mode,
  tui,
};

pub struct App {
  pub config: Config,
  pub clients: NonEmpty<matrix_sdk::Client>,
  pub commands: RataCommands,
  pub tick_rate: f64,
  pub frame_rate: f64,
  pub components: Vec<Box<dyn Component>>,
  pub should_quit: AtomicBool,
  pub should_suspend: bool,
  pub mode: Mode,
  pub last_tick_key_events: RefCell<Vec<KeyEvent>>,
}

impl App {
  pub async fn new(tick_rate: f64, frame_rate: f64) -> Result<Self> {
    let home = Home::new();
    let fps = FpsCounter::default();
    let config = Config::new()?;
    let mode = Mode::Home;
    let future_clients = config.accounts.clone().map(|conf| {
      tokio::spawn(async move {
        let server_name = conf.user_id.server_name();
        let client = matrix_sdk::Client::builder()
          .server_name(server_name)
          .build()
          .await
          .with_context(|| format!("Could not initialize client for {}", server_name))?;
        client
          .matrix_auth()
          .login_username(&conf.user_id, &conf.password)
          .initial_device_display_name(&conf.device_name)
          .send()
          .await
          .with_context(|| format!("Could not login as {}", conf.user_id))?;
        Ok::<_, color_eyre::eyre::Report>(client)
      })
    });
    let mut clients = Vec::new();
    for client in futures::future::join_all(future_clients).await {
      clients.push(
        client
          .context("Could not join client init task")?
          .context("Could not connect to all accounts")?,
      );
    }
    let clients = NonEmpty::collect(clients).expect("map on NonEmpty returned empty vec");

    let mut app = Self {
      tick_rate,
      frame_rate,
      components: vec![Box::new(home), Box::new(fps)],
      should_quit: AtomicBool::new(false),
      should_suspend: false,
      clients,
      config,
      commands: RataCommands::new(),
      mode,
      last_tick_key_events: RefCell::new(Vec::new()),
    };
    crate::plugins::load_all(&mut app).context("Could not load plugins")?;
    Ok(app)
  }

  pub async fn run(&mut self) -> Result<()> {
    let (action_tx, mut action_rx) = mpsc::unbounded_channel();
    let (sync_responses_tx, mut sync_responses_rx) = mpsc::unbounded_channel();

    let mut tui = tui::Tui::new()?
      .tick_rate(self.tick_rate)
      .frame_rate(self.frame_rate);
    // tui.mouse(true);
    tui.enter()?;

    for component in self.components.iter_mut() {
      component.register_action_handler(action_tx.clone())?;
    }

    for component in self.components.iter_mut() {
      component.register_config_handler(self.config.clone())?;
    }

    for component in self.components.iter_mut() {
      component.init(tui.size()?)?;
    }

    let sync_clients = self.clients.clone();
    let mut sync_results = FuturesUnordered::new();
    for client in sync_clients.iter() {
      client.add_event_handler(
        |ev: matrix_sdk::ruma::events::room::message::SyncRoomMessageEvent| async move {
          println!("Received a message {:?}", ev);
        },
      );
      let server_name = client.user_id().expect("missing user id").server_name();
      let mut sync_settings = matrix_sdk::config::SyncSettings::default();
      if server_name == "matrix.org" {
        // matrix.org is slow and frequently hits the timeout on initial sync,
        // let's quadruple the default
        sync_settings = sync_settings.timeout(std::time::Duration::from_secs(120));
      }
      sync_results.push(client.sync_with_result_callback(sync_settings, |res| {
        let client2 = client.clone();
        async {
          let sync_response = res.expect("Failed sync");
          sync_responses_tx
            .send((client2, sync_response))
            .expect("could not send sync response");
          if self.should_quit.load(Ordering::Acquire) {
            Ok(matrix_sdk::LoopCtrl::Break)
          } else {
            Ok(matrix_sdk::LoopCtrl::Continue)
          }
        }
      }));
    }

    loop {
      tokio::select! {
        e = tui.next() => {
          if let Some(e) = e {
            self.handle_tui_event(&action_tx, e.clone()).context("Error while handling TUI event")?;
            for component in self.components.iter_mut() {
              if let Some(action) = component.handle_events(Some(e.clone()))? {
                action_tx.send(action).context("Error while sending action from component")?;
              }
            }
          }
        }
        sync_response = sync_responses_rx.recv() => {
          let (client, sync_response) = sync_response.expect("sync_responses_rx unexpectedly closed");
          self.handle_sync_response(&action_tx, client, sync_response).await.context("Error while handling sync response")?;
        }
        sync_result = sync_results.next() => {
          if !self.should_quit.load(Ordering::Acquire) {
            panic!("Sync ended unexpected: {:?}", sync_result);
          }
        }
      }

      while let Ok(action) = action_rx.try_recv() {
        if action != Action::Tick && action != Action::Render {
          log::debug!("{action:?}");
        }
        match action {
          Action::Tick => {
            self.last_tick_key_events.borrow_mut().drain(..);
          },
          Action::Quit => self.should_quit.store(true, Ordering::Release),
          Action::Suspend => self.should_suspend = true,
          Action::Resume => self.should_suspend = false,
          Action::Resize(w, h) => {
            tui.resize(Rect::new(0, 0, w, h))?;
            tui.draw(|f| {
              for component in self.components.iter_mut() {
                let r = component.draw(f, f.size());
                if let Err(e) = r {
                  action_tx
                    .send(Action::Error(format!("Failed to draw: {:?}", e)))
                    .unwrap();
                }
              }
            })?;
          },
          Action::Render => {
            tui.draw(|f| {
              for component in self.components.iter_mut() {
                let r = component.draw(f, f.size());
                if let Err(e) = r {
                  action_tx
                    .send(Action::Error(format!("Failed to draw: {:?}", e)))
                    .unwrap();
                }
              }
            })?;
          },
          _ => {},
        }
        for component in self.components.iter_mut() {
          if let Some(action) = component.update(action.clone())? {
            action_tx.send(action)?
          };
        }
      }
      if self.should_suspend {
        tui.suspend()?;
        action_tx.send(Action::Resume)?;
        tui = tui::Tui::new()?
          .tick_rate(self.tick_rate)
          .frame_rate(self.frame_rate);
        // tui.mouse(true);
        tui.enter()?;
      } else if self.should_quit.load(Ordering::Acquire) {
        tui.stop()?;
        break;
      }
    }
    tui.exit()?;
    Ok(())
  }

  fn handle_tui_event(
    &self,
    action_tx: &mpsc::UnboundedSender<Action>,
    e: tui::Event,
  ) -> Result<()> {
    match e {
      tui::Event::Quit => action_tx.send(Action::Quit)?,
      tui::Event::Tick => action_tx.send(Action::Tick)?,
      tui::Event::Render => action_tx.send(Action::Render)?,
      tui::Event::Resize(x, y) => action_tx.send(Action::Resize(x, y))?,
      tui::Event::Key(key) => {
        if let Some(keymap) = self.config.keybindings.get(&self.mode) {
          if let Some(command_line) = keymap.get(&vec![key]) {
            log::info!("Got command: {command_line}");
            crate::commands::run_command(command_line, &self, &action_tx)?;
          } else {
            let mut last_tick_key_events = self.last_tick_key_events.borrow_mut();

            // If the key was not handled as a single key action,
            // then consider it for multi-key combinations.
            last_tick_key_events.push(key);

            // Check for multi-key combinations
            if let Some(command_line) = keymap.get(&*last_tick_key_events) {
              log::info!("Got command: {command_line}");
              crate::commands::run_command(command_line, &self, &action_tx)?;
            }
          }
        };
      },
      _ => {},
    }

    Ok(())
  }

  async fn handle_sync_response(
    &self,
    action_tx: &mpsc::UnboundedSender<Action>,
    client: matrix_sdk::Client,
    sync_response: matrix_sdk::sync::SyncResponse,
  ) -> Result<()> {
    println!("{:?}", sync_response);
    Ok(())
  }
}
